/*
ME333 Assignment 5 Part 2
Track desired signal with LED/phototransistor. 

2/18/11 Changed to use AN0 & changed initial gains to Andrew's HW gain values -NR
2/2011 Andrew Kessler
*/

#include <plib.h>
#include "NU32v2.h"
#include "Nokia5110.h"

#define SAMPLES 1000 //size of the arrays
#define CTRL_FREQ 1000 //control loop freq (Hz)
#define REF_FREQ 2.0
#define REF_DUTY 0.5
#define V_LO 1
#define V_HI 2
#define LIMIT_INT 40000000 //as big as the integral error is allowed to go
#define SAMPLE_TIME 8


/*Function Prototypes:*/
void wait();
void initADC();
void createSignal();
void initPWM();
void initCtrlLoop();
void LcdStartUp();

//Global variables:

float ADCvalues[SAMPLES]; //1000 adc values
float desiredVs[SAMPLES]; //1000 desired voltage values
int dutyCycles[SAMPLES];//1000 pwm duty cycles
int global_i; //index for the arrays
float kp,ki,error,DCprop,errorInt,DCint; //stuff for control loop
int ADCtemp,DCout;
char lcdBuffer[100];

int main(void) {

  SYSTEMConfig(SYS_FREQ, SYS_CFG_ALL);

  // configure for multi-vectored mode
  INTConfigureSystem(INT_SYSTEM_CONFIG_MULT_VECTOR);

  // enable multi-vector interrupts 
  INTEnableSystemMultiVectoredInt(); 

  
  initSerialNU32v2(); // initialize serial port and interrupt
  initLEDs(); // set LED pins as outputs
  ki=0.02;//Initialize the gains to 0
  kp=0.06;
  LcdStartUp(); //intialize LCD and print basic information
  
  global_i=0;  //Initialize the index to 0
  errorInt=0;
  createSignal(); //fill in desired voltage array
  initADC();//Initialize the adc to read pin B0  
  initPWM(); //Initialize the pwm to 100kHz with a duty cycle between 0 and 800
  initCtrlLoop(); //Initialize a timer to interrupt every 1ms

  while(1) {
    // nothing, do everything happens in interupts
  }

  return 0;
} // end main

/* ISR for Timer 4 (Control Loop)*/
void __ISR(_TIMER_4_VECTOR, ipl2) Timer4Handler(void){
	int i,v;
	//read the analog signal
	AD1CON1bits.SAMP =1; //start sampling
	WriteTimer3(1); //reset the ADC timer
  	while(ReadTimer3() <= SAMPLE_TIME){}; //sample for 200ns (> 132 ns)
  	AD1CON1bits.SAMP= 0; //stop sampling and start converstion to a 10 bit number
  	while(!AD1CON1bits.DONE){}; //wait here unil conversion finishes
  	ADCtemp = ADC1BUF0; //read the buffer location where the result is stored
  	
	ADCvalues[global_i]=ADCtemp; //save it into the buffer at the index position	
  	
  	//calculate the error
    error=((desiredVs[global_i]*800.0)/3.3)-((ADCtemp*800.0)/1024.0); //error is in terms of DC values 0-800
  	DCprop=kp*error; //proportional controller
  	errorInt+=error;
  	if (global_i==0||errorInt>LIMIT_INT) //reset the errorInt after every new signal, or if its too high
  		{errorInt=0;}
	DCint=ki*errorInt; //integral controller
    
	DCout=(int) DCprop + DCint;
	//make sure between 0 and 800
	if (DCout>799) 
		{DCout=799;}
	if (DCout<0)
		{DCout=0;}
    dutyCycles[global_i]=DCout; //store in buffer
    SetDCOC1PWM(DCout); //set the duty cycle to the desired output

	if (global_i%100==0){ //update LCD every 100 counts
	gotoXY(0,3);
	sprintf(lcdBuffer, "Error=%4.2f", error);
  	LcdString(lcdBuffer);
	gotoXY(0,4);
	sprintf(lcdBuffer, "DCout=%d", DCout);
  	LcdString(lcdBuffer);
 	} 	

    global_i++; //increment counter
    
    if (global_i>999){
    	global_i=0;
    	CloseTimer4(); //DisableIntT4(); //turn off timer interupt
    	gotoXY(0,5);
  		LcdString("Writing...");
    	
  		//send the 1000 adc values to the plotting tool trace a
  		 for (i=0; i<1000; i++) {
	  		 v = (int) (ADCvalues[i]/2.0);
	  		 sprintf(RS232_Out_Buffer, "a %d %d\r\n", i, v); // max adc is 1023, max plotted value is 511
      	     WriteString(UART2A, RS232_Out_Buffer);
     		} 	     
		//send the 1000 pwm duty cycle values to plotting tool trace b
		for (i=0; i<1000; i++) {
			 v = (int) (dutyCycles[i]*511.0)/800.0;
	  		 sprintf(RS232_Out_Buffer, "b %d %d\r\n", i, v); // max duty is 800, max plotted value is 511
      	     WriteString(UART2A, RS232_Out_Buffer);
     		}
  		//send the 1000 desired voltage to the plotting tool trace c
  		for (i=0; i<1000; i++) {
	  	v= (int)  desiredVs[i]*(511.0/3.3);
	  	sprintf(RS232_Out_Buffer, "c %d %d\r\n", i,v); // max desiredV is 3.3, max plotted value is 511
        WriteString(UART2A, RS232_Out_Buffer);
     		}
  		//send the value of your control gains to the plotting tool
  		sprintf(RS232_Out_Buffer, " kp=%4.2f\r\n ki=%4.2f\r\n", kp,ki);
        WriteString(UART2A, RS232_Out_Buffer);
  		wait();//delay 0.5s
  		
  		initCtrlLoop(); //turn timer back on
  		gotoXY(0,5);
  		LcdString("Running..."); //indicate that control law is operating again
		
  
  } 
  	mT4ClearIntFlag(); //clear the interrupt flag
}
/*
UART Interrupt
-if the NU32v2 receives a 'r', increase kp
-if the NU32v2 receives a 't', decrease kp

-if the NU32v2 receives a 'f', increase ki
-if the NU32v2 receives a 'g', decrease ki
*/ 

void __ISR(_UART_2A_VECTOR, ipl2) IntUart2AHandler(void) {
  // Is this an RX interrupt?
  if(INTGetFlag(INT_SOURCE_UART_RX(UART2A))){
    char data = UARTGetDataByte(UART2A);

    // now do something with data
    if (data == 'r') { // increase kp
      kp=kp+0.01;
    }
    if (data == 't') { // descrease kp
      kp=kp-0.01;
    }
    if (data=='y')
    	kp=kp+0.1;
    if (data == 'f') { // increase ki
      ki=ki+0.01;
    }
    if (data == 'g') { // descrease ki
      ki=ki-0.01;
    }
    gotoXY(0,1);
    sprintf(lcdBuffer, " kp=%4.2f", kp);
  	LcdString(lcdBuffer);
  	gotoXY(0,2);
	sprintf(lcdBuffer, " ki=%4.2f", ki);
  	LcdString(lcdBuffer);
    // Clear the RX interrupt Flag
    INTClearFlag(INT_SOURCE_UART_RX(UART2A)); 
  }

  // We don't care about TX interrupt
  if(INTGetFlag(INT_SOURCE_UART_TX(UART2A))) {
    INTClearFlag(INT_SOURCE_UART_TX(UART2A));
  }
}
void initADC(){
//set up ADC to read from AN15 and store the sample, so we can read it using int
//AD1PCG = 0xFFFD; //set B0 to analog input
AD1CON1bits.FORM = 4; //set output format to range from 0-1023
AD1CHSbits.CH0SA = 0; //connect pin AN15 to MUXA for sampling
AD1CON3bits.ADCS = 3; //set conversion period = 2(ADCS+1)*PBClk
AD1CON1bits.ADON = 1; //turn on A/D converter

//use TMR3 to ensure that we sample the input for at least 132ns
OpenTimer3(T3_ON | T3_PS_1_2, 65535);	
}	
void createSignal(){
	//Initialize the desired voltage values to a 2Hz square wave between 1V and 2V
  float tHigh,tLow;
  int indicesHigh,indicesLow,startIndex,i,j;
 
  tHigh=REF_DUTY/REF_FREQ; //period signal is high
  tLow=1/REF_FREQ - tHigh; //period signal is low
  indicesHigh=tHigh*SAMPLES; //the number of indices signal is high
  indicesLow= tLow*SAMPLES; //the number of indices the signal is low
  startIndex=0; //keeps track of where the next entry should be
  
  //SAMPLES*REF_FREQ/CTRL_FREQ
  for (j=0; j<2; j++){ //repeat outer loop for number of signal/sec (2Hz= 2x)
	  
	  //fill the low Vs
	  for (i=0; i<indicesLow; i++) 
	  	{desiredVs[i+startIndex]=V_LO;
	  	}
	  startIndex+=i;
	  
	  //fill the high Vs
	  for (i=0; i<indicesHigh; i++)
		  {desiredVs[i+startIndex]=V_HI;}
	  startIndex+=i; 
	 } 
	 
}	 
void initPWM(){ 
	OpenTimer2(T2_ON | T2_PS_1_1, 800); // 100kHz PWM: 80000000Hz / 1ps / 800 = 100 kHz
	OpenOC1(OC_ON | OC_TIMER2_SRC | OC_PWM_FAULT_PIN_DISABLE, 0, 0); // PWM output on Pin D0, 0 duty cycle
}
void initCtrlLoop(){
	int ctrfrq=SYS_FREQ/CTRL_FREQ;
	OpenTimer4(T4_ON | T4_SOURCE_INT | T4_PS_1_1, ctrfrq); //interupt at CTRL_FREQ (1 KHz)
	ConfigIntTimer4(T4_INT_ON | T4_INT_PRIOR_2);

}	

/* Delay function - Pause for 0.5s */
void wait(){
int k;
for(k=0; k<40000000; k++){}; //.5*SYS_FREQ = 40e6	
}	
void LcdStartUp(){
  LcdInitialize();
  LcdClear();
  gotoXY(0,0);
  LcdString("Gains:");
  gotoXY(0,1);
  sprintf(lcdBuffer, " kp=%4.2f", kp);
  LcdString(lcdBuffer);
  gotoXY(0,2);
  sprintf(lcdBuffer, " ki=%4.2f", ki);
  LcdString(lcdBuffer);
}  
